using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

namespace Sturm.Util.ApplicationServices.ObjectPooling {
	/// <summary>
	/// This is the main Pool class, which implements object pooling.
	/// </summary>
	public sealed class Pool<T> {
		#region Slot class
		/// <summary>
		/// The slot class represents an entry in the pool's list. It holds information about
		/// the actual pooled object, plus per-object management information.
		/// </summary>
		sealed class Slot {
			/// <summary>
			/// Constructs a Slot object for a pre-existing pooled type object.
			/// </summary>
			public Slot(T poolObject) {
				if (poolObject == null)
					throw new ArgumentNullException("poolObject");
				this.poolObject = poolObject;
			}

			/// <summary>
			/// Constructs a Slot object without a pre-existing pooled type object.
			/// Creation of the pooled type object will be deferred to the point
			/// where the object is actually accessed via the PoolObject property.
			/// </summary>
			public Slot(IObjectFactory<T> objectFactory) {
				if (objectFactory == null)
					throw new ArgumentNullException("objectFactory");
				this.objectFactory = objectFactory;
			}

			bool inUse;
			/// <summary>
			/// Returns a value indicating whether this slot is currently in use or not, in other words
			/// whether its object has been reserved for use at this time.
			/// </summary>
			public bool InUse {
				get {
					return inUse;
				}
			}

			/// <summary>
			/// Marks the slot as being in use.
			/// </summary>
			public void Occupy( ) {
				inUse = true;
			}

			/// <summary>
			/// Marks the slot as being available.
			/// </summary>
			public void Release( ) {
				inUse = false;
			}

			IObjectFactory<T> objectFactory;
			T poolObject;
			/// <summary>
			/// Returns the pooled object itself. If there isn't any object yet, 
			/// it'll be created on the fly using the object factory.
			/// </summary>
			public T PoolObject {
				get {
					if (poolObject == null)
						poolObject = objectFactory.CreateObject( );
					return poolObject;
				}
			}
		}
		#endregion

 		#region Constructors
		private Pool( ) {
			pool = new List<Slot>( );
			retryWaitTime = 500;
			poolExtensionBatchSize = 10;
			poolExtensionBatchSizeIsPercent = true;
			maxGetObjectTries = 5;
			outOfObjectsBehaviour = OutOfObjectsBehaviour.ExtendPool;
			delayObjectCreation = false;
			maxPoolSize = 100;
			lowWaterMark = 20;
			highWaterMark = 80;
			resizingWaitTime = 2000;
			growThreshold = 0;
			shrinkThreshold = 3;
		}

		/// <summary>
		/// Creates a new Pool object. The list of objects in the pool is 
		/// expanded by a number of additional slots, as given by the
		/// initialPoolSize parameter. Objects are created by a call to the
		/// IObjectFactory&lt;T&gt;.CreateObject() method.
		/// </summary>
		public Pool(IObjectFactory<T> objectFactory, int initialPoolSize)
			: this( ) {
			this.objectFactory = objectFactory;
			if (initialPoolSize > 0)
				ExtendPoolBy(initialPoolSize);
		}

		/// <summary>
		/// Constructs a pool and fills its list of pooled objects with the objects
		/// passed in to this constructor.
		/// </summary>
		public Pool(IEnumerable<T> objects)
			: this( ) {
			foreach (T entry in objects)
				pool.Add(new Slot(entry));
		}
		#endregion

		#region Properties
		private OutOfObjectsBehaviour outOfObjectsBehaviour;
		/// <summary>
		/// Gets or sets a value indicating the behaviour that the GetObject()
		/// method shows when a pool object isn't (immediately) available.
		/// </summary>
		public OutOfObjectsBehaviour OutOfObjectsBehaviour {
			get {
				return outOfObjectsBehaviour;
			}
			set {
				if (outOfObjectsBehaviour != value) {
					outOfObjectsBehaviour = value;
				}
			}
		}

		private int retryWaitTime;
		/// <summary>
		/// Gets or sets a value that defines the time to wait between tries
		/// to acquire a pool object in the GetObject() method. This is only 
		/// evaluated for OutOfObjectsBehaviour == WaitAndRetry.
		/// </summary>
		public int RetryWaitTime {
			get {
				return retryWaitTime;
			}
			set {
				if (retryWaitTime != value) {
					retryWaitTime = value;
				}
			}
		}

		private IObjectFactory<T> objectFactory;
		/// <summary>
		/// Gets or sets a value defining the object factory that can be used 
		/// by the pool to create new poolable objects as necessary.
		/// </summary>
		public IObjectFactory<T> ObjectFactory {
			get {
				return objectFactory;
			}
			set {
				if (objectFactory != value) {
					objectFactory = value;
				}
			}
		}

		private int poolExtensionBatchSize;
		/// <summary>
		/// Gets or sets a value that defines the size of the batch when the pool is 
		///  growing. This can be an absolute value or a percentage, which is 
		/// defined by the PoolExtensionBatchSizeIsPercent property.
		/// </summary>
		public int PoolExtensionBatchSize {
			get {
				return poolExtensionBatchSize;
			}
			set {
				if (poolExtensionBatchSize != value) {
					poolExtensionBatchSize = value;
				}
			}
		}

		private bool poolExtensionBatchSizeIsPercent;
		/// <summary>
		/// Gets or sets a flag that indicates whether the value given by 
		/// the PoolExtensionBatchSize property is an absolute value or a percentage.
		/// </summary>
		public bool PoolExtensionBatchSizeIsPercent {
			get {
				return poolExtensionBatchSizeIsPercent;
			}
			set {
				if (poolExtensionBatchSizeIsPercent != value) {
					poolExtensionBatchSizeIsPercent = value;
				}
			}
		}

		private int maxGetObjectTries;
		/// <summary>
		/// Gets or sets a value that defines the maximum tries that are made 
		/// to acquire an object from the pool in the GetObject() method.
		/// </summary>
		public int MaxGetObjectTries {
			get {
				return maxGetObjectTries;
			}
			set {
				if (maxGetObjectTries != value) {
					maxGetObjectTries = value;
				}
			}
		}

		private bool delayObjectCreation;
		/// <summary>
		/// Gets or sets a flag that defines whether a poolable type object is 
		/// immediately created together with a new Slot, or whether the 
		/// creation of the poolable type object will be delayed.
		/// </summary>
		public bool DelayObjectCreation {
			get {
				return delayObjectCreation;
			}
			set {
				if (delayObjectCreation != value) {
					delayObjectCreation = value;
				}
			}
		}

		private int maxPoolSize;
		/// <summary>
		/// Gets or sets a value indicating the maximum number of objects that the pool
		/// shall hold. The ExtendPoolBy() method won't extend the pool to a larger
		/// size than this.
		/// </summary>
		public int MaxPoolSize {
			get {
				return maxPoolSize;
			}
			set {
				if (maxPoolSize != value) {
					maxPoolSize = value;
					lock (poolLock) {
						if (pool.Count > maxPoolSize)
							ShrinkPoolBy(pool.Count - maxPoolSize);
					}
				}
			}
		}

		Timer resizingTimer;
		object resizingTimerLock = new object( );

		private bool useTimerBasedResizing;
		/// <summary>
		/// Gets or sets a value indicating whether the pool checks for low and high
		/// water situations automatically and regularly. Also see ResizingWaitTime.
		/// </summary>
		public bool UseTimerBasedResizing {
			get {
				return useTimerBasedResizing;
			}
			set {
				if (useTimerBasedResizing != value) {
					useTimerBasedResizing = value;
					lock (resizingTimerLock) {
						if (useTimerBasedResizing == false && resizingTimer != null)
							resizingTimer.Dispose( );
						else if (useTimerBasedResizing == true)
							resizingTimer = new Timer(new TimerCallback(ResizingTimerCallback), null, resizingWaitTime, Timeout.Infinite);
					}
				}
			}
		}

		private int resizingWaitTime;
		/// <summary>
		/// Gets or sets a value that indicates the time between to checks for low and 
		/// high water situations. Also see UseTimerBasedResizing.
		/// </summary>
		public int ResizingWaitTime {
			get {
				return resizingWaitTime;
			}
			set {
				if (resizingWaitTime != value) {
					resizingWaitTime = value;
					lock (resizingTimerLock) {
						if (resizingTimer != null)
							resizingTimer.Change(resizingWaitTime, Timeout.Infinite);
					}
				}
			}
		}

		private int highWaterMark;
		/// <summary>
		/// Gets or sets a value indicating the percentage of used objects in the pool
		/// that must be exceeded for the pool to be grown automatically.
		/// </summary>
		public int HighWaterMark {
			get {
				return highWaterMark;
			}
			set {
				if (highWaterMark != value) {
					highWaterMark = value;
				}
			}
		}

		private int lowWaterMark;
		/// <summary>
		/// Gets or sets a value that the percentage of used objects in the pool
		/// must fall below for the pool to be shrunken automatically.
		/// </summary>
		public int LowWaterMark {
			get {
				return lowWaterMark;
			}
			set {
				if (lowWaterMark != value) {
					lowWaterMark = value;
				}
			}
		}

		private int shrinkThreshold;
		/// <summary>
		/// Gets or sets a value indicating the number of times the pool must 
		/// be judged too big before it is shrunken.
		/// </summary>
		public int ShrinkThreshold {
			get {
				return shrinkThreshold;
			}
			set {
				if (shrinkThreshold != value) {
					shrinkThreshold = value;
				}
			}
		}

		private int growThreshold;
		/// <summary>
		/// Gets or sets a value indicating the number of times the pool must 
		/// be judged too small before it is grown.
		/// </summary>
		public int GrowThreshold {
			get {
				return growThreshold;
			}
			set {
				if (growThreshold != value) {
					growThreshold = value;
				}
			}
		}

		/// <summary>
		/// Gets a value indicating the current size of the pool.
		/// </summary>
		public int Size {
			get {
				lock(poolLock)
					return pool.Count;
			}
		}

		private int inUseCount;
		/// <summary>
		/// Gets a value indicating the current number of used objects in the pool.
		/// </summary>
		public int InUseCount {
			get {
				return inUseCount;
			}
		}

		/// <summary>
		/// Gets a value indicating the percentage of used objects in the pool.
		/// </summary>
		public float  FillLevel {
			get {
				return InUseCount * 100f / Size;
			}
		}
		#endregion

		#region Pool access

		List<Slot> pool;
		object poolLock = new object( );

		/// <summary>
		/// Returns a pooled object. After use, the pooled object must be returned to the
		/// pool by calling the ReleaseObject method.
		/// </summary>
		public T GetObject( ) {
			int tryNo = 1;

			while (tryNo <= maxGetObjectTries) {
				lock (poolLock) {
					Slot unusedSlot = FindUnusedSlot( );
					if (unusedSlot != null) {
						unusedSlot.Occupy( );
						inUseCount++;
						return unusedSlot.PoolObject;
					}
				}

				switch (outOfObjectsBehaviour) {
					case OutOfObjectsBehaviour.ReturnNull:
						return default(T);
					case OutOfObjectsBehaviour.ThrowException:
						throw new OutOfObjectsException( );
					case OutOfObjectsBehaviour.WaitAndRetry:
						Thread.Sleep(retryWaitTime);
						break;
					case OutOfObjectsBehaviour.ExtendPool:
						ExtendPool( );
						break;
				}
			}

			throw new OutOfObjectsException( );
		}

		Slot FindUnusedSlot( ) {
			return pool.Find(delegate(Slot slot) {
				return !slot.InUse;
			});
		}

		/// <summary>
		/// Releases an object that was previously fetched from the pool
		/// by a call to the GetObject method.
		/// </summary>
		public void ReleaseObject(T poolObject) {
			lock (poolLock) {
				Slot objectSlot = pool.Find(delegate(Slot slot) {
					return EqualityComparer<T>.Default.Equals(slot.PoolObject, poolObject);
				});
				if (objectSlot != null) {
					objectSlot.Release( );
					inUseCount--;
				}
				else
					throw new ArgumentException("The given object is not part of this pool.");
			}
		}
		#endregion

		#region Size handling

		/// <summary>
		/// Extends the pool by a number of objects that's specified by the PoolExtensionBatchSize
		/// and the PoolExtensionBatchSizeIsPercent properties.
		/// </summary>
		void ExtendPool( ) {
			lock (poolLock) {
				int newObjects = poolExtensionBatchSizeIsPercent ? pool.Count * poolExtensionBatchSize / 100 : poolExtensionBatchSize;
				ExtendPoolBy(Math.Max(newObjects, 1));
			}
		}

		/// <summary>
		/// Extends the number of slots in the pool by the number given in
		/// the count parameter. 
		/// </summary>
		public void ExtendPoolBy(int count) {
			if (objectFactory == null)
				throw new InvalidOperationException("An ObjectFactory must be configured.");
			if (count <= 0)
				throw new ArgumentOutOfRangeException("Count must be greater than zero.");
			lock (poolLock) {
				count = Math.Min(maxPoolSize - pool.Count, count);

				for (int i = 0; i < count; i++) {
					Slot slot = delayObjectCreation ? new Slot(objectFactory) :
						new Slot(objectFactory.CreateObject( ));
					pool.Add(slot);
				}
			}
		}

		/// <summary>
		/// Shrinks the pool size by count elements, at most. The method removes
		/// only unused pool objects, as long as it finds them.
		/// </summary>
		public void ShrinkPoolBy(int count) {
			if (count <= 0)
				throw new ArgumentOutOfRangeException("Count must be greater than zero.");
			lock (poolLock) {
				count = Math.Max(pool.Count - count, 0);

				Slot unusedSlot;
				do {
					unusedSlot = FindUnusedSlot( );
					if (unusedSlot != null)
						pool.Remove(unusedSlot);
					count--;
				} while (count > 0 && unusedSlot != null);
			}
		}

		int fillLevelFoundTooLow;
		int fillLevelFoundTooHigh;
		int inUseCountSum;

		private void CheckFillLevel(float fillLevel, int objectsInUse) {
			if (fillLevel < lowWaterMark) {
				fillLevelFoundTooLow++;
				if (fillLevelFoundTooHigh > 0) {
					// At the last check, the fill level was too high
					fillLevelFoundTooHigh = 0;
					inUseCountSum = objectsInUse;
				}
				else
					inUseCountSum += objectsInUse;
			}
			else if (fillLevel > highWaterMark) {
				fillLevelFoundTooHigh++;
				if (fillLevelFoundTooLow > 0) {
					// At the last check, the fill level was too low
					fillLevelFoundTooLow = 0;
					inUseCountSum = objectsInUse;
				}
				else
					inUseCountSum += objectsInUse;
			}
			else
				fillLevelFoundTooHigh = fillLevelFoundTooLow = 0;
		}

		private void GrowOrShrink(float fillLevel, int objectsInUse, int currentSize) {
			if (fillLevelFoundTooLow > shrinkThreshold) {
				// We have found the pool to be too large for a number of times (in a row) greater than
				// shrinkThreshold. For now, we shrink the pool to a number of objects, so that the
				// average used objects count over the last x runs is no longer below the
				// LowWaterMark, where x is fillLevelFoundTooLow.
				int targetSize = inUseCountSum / fillLevelFoundTooLow;
				int targetPoolSize = (int) Math.Round(targetSize * 100 / LowWaterMark + 0.5);

				ShrinkPoolBy(currentSize - targetPoolSize);

				fillLevelFoundTooLow = 0;
				inUseCountSum = 0;
			}
			else if (fillLevelFoundTooHigh > growThreshold) {
				// We have found the pool to be too small for a number of times (in a row) greater than
				// growThreshold. We want to grow the pool. The target is to bring the pool size to a number
				// so that the current used objects count is no longer above the HighWaterMark.
				// At the same time, we want to take into account the used object numbers from the
				// other recent runs where the fill level was also above the HighWaterMark.
				// To do this, we calculate the average number of used objects over these runs, but
				// give the last run the highest priority. Then the resize number is calculated so that the
				// given target size equals HighWaterMark % of the pool size.
				int averageUsedObjects = inUseCountSum / fillLevelFoundTooHigh;
				int targetSize = Math.Max(averageUsedObjects, objectsInUse);
				int targetPoolSize = (int) Math.Round(targetSize * 100 / HighWaterMark + 0.5);

				ExtendPoolBy(targetPoolSize - currentSize);

				fillLevelFoundTooHigh = 0;
				inUseCountSum = 0;
			}
		}

		void ResizingTimerCallback(object state) {
			int objectsInUse = InUseCount;
			int currentSize = Size;
			// I calculate the fill level myself, so I can be sure it's consistent with
			// the other two values.
			float fillLevel = objectsInUse * 100f / currentSize;

			CheckFillLevel(fillLevel, objectsInUse);
			GrowOrShrink(fillLevel, objectsInUse, currentSize);

			lock (resizingTimerLock)
				if (resizingTimer != null)
					resizingTimer.Change(resizingWaitTime, Timeout.Infinite);
		}
		#endregion
	}
}
